from collections.abc import Callable
from typing import Any

from django.db.models import Model
from django.forms import Form
from django.http import HttpRequest
from django.template.response import TemplateResponse
from django.urls import reverse

from unfold.dataclasses import UnfoldAction
from unfold.enums import ActionVariant
from unfold.exceptions import UnfoldException


class ActionModelAdminMixin:
    """
    Adds support for various ModelAdmin actions (list, detail, row, submit line)
    """

    actions_list = ()  # Displayed in changelist at the top
    actions_row = ()  # Displayed in changelist for each row in the table
    actions_detail = ()  # Displayed in changeform at the top
    actions_submit_line = ()  # Displayed in changeform in the submit line (form buttons)

    def changelist_view(
        self, request: HttpRequest, extra_context: dict[str, str] | None = None
    ) -> TemplateResponse:
        """
        Changelist contains `actions_list` and `actions_row` custom actions. In case of `actions_row` they
        are displayed in the each row of the table.
        """
        extra_context = extra_context or {}

        actions_row = [
            {
                "title": action.description,
                "icon": action.icon,
                "attrs": action.method.attrs,
                # This is just a path name as string and in template is used in {% url %} tag
                # with custom instance pk value
                "raw_path": f"{self.admin_site.name}:{action.action_name}",
            }
            for action in self.get_actions_row(request)
        ]

        # `actions_list` may contain custom structure with dropdowns so it is needed
        # to use `_get_actions_navigation` to build the final structure for the template
        actions_list = self._get_actions_navigation(
            self.actions_list, self.get_actions_list(request)
        )

        extra_context.update(
            {
                "actions_list": actions_list,
                "actions_row": actions_row,
            }
        )

        return super().changelist_view(request, extra_context)

    def changeform_view(
        self,
        request: HttpRequest,
        object_id: str | None = None,
        form_url: str = "",
        extra_context: dict[str, Any] | None = None,
    ) -> Any:
        """
        Changeform contains `actions_submit_line` and `actions_detail` custom actions.
        """
        extra_context = extra_context or {}

        if object_id:
            # `actions_submit_line` is a list of actions that are displayed in the submit line they
            # are displayed as form buttons
            actions_submit_line = self.get_actions_submit_line(request, object_id)

            # `actions_detail` may contain custom structure with dropdowns so it is needed
            # to use `_get_actions_navigation` to build the final structure for the template
            actions_detail = self._get_actions_navigation(
                self.actions_detail,
                self.get_actions_detail(request, object_id),
                object_id,
            )

            extra_context.update(
                {
                    "actions_submit_line": actions_submit_line,
                    "actions_detail": actions_detail,
                }
            )

        return super().changeform_view(request, object_id, form_url, extra_context)

    def save_model(
        self, request: HttpRequest, obj: Model, form: Form, change: Any
    ) -> None:
        """
        When saving object, run all appropriate actions from `actions_submit_line`
        """
        super().save_model(request, obj, form, change)

        # After saving object, check if any button from `actions_submit_line` was pressed
        # and call the corresponding method
        for action in self.get_actions_submit_line(request, obj.pk):
            if action.action_name not in request.POST:
                continue

            action.method(request, obj)

    def get_unfold_action(self, action: str) -> UnfoldAction:
        """
        Converts action name into UnfoldAction object.
        """
        method = self._get_instance_method(action)

        return UnfoldAction(
            action_name=f"{self.model._meta.app_label}_{self.model._meta.model_name}_{action}",
            method=method,
            description=self._get_action_description(method, action),
            path=getattr(method, "url_path", action),
            attrs=method.attrs if hasattr(method, "attrs") else None,
            icon=method.icon if hasattr(method, "icon") else None,
            variant=method.variant if hasattr(method, "variant") else None,
        )

    def get_actions_list(self, request: HttpRequest) -> list[UnfoldAction]:
        """
        Filters `actions_list` by permissions and returns list of UnfoldAction objects.
        """
        return self._filter_unfold_actions_by_permissions(
            request, self._get_base_actions_list()
        )

    def get_actions_detail(
        self, request: HttpRequest, object_id: int
    ) -> list[UnfoldAction]:
        """
        Filters `actions_detail` by permissions and returns list of UnfoldAction objects.
        """
        return self._filter_unfold_actions_by_permissions(
            request, self._get_base_actions_detail(), object_id
        )

    def get_actions_row(self, request: HttpRequest) -> list[UnfoldAction]:
        """
        Filters `actions_row` by permissions and returns list of UnfoldAction objects.
        """
        return self._filter_unfold_actions_by_permissions(
            request, self._get_base_actions_row()
        )

    def get_actions_submit_line(
        self, request: HttpRequest, object_id: int
    ) -> list[UnfoldAction]:
        """
        Filters `actions_submit_line` by permissions and returns list of UnfoldAction objects.
        """
        return self._filter_unfold_actions_by_permissions(
            request, self._get_base_actions_submit_line(), object_id
        )

    def _extract_action_names(self, actions: list[str | dict]) -> list[str]:
        """
        Gets the list of only actions names from the actions structure provided in ModelAdmin
        """
        results = []

        for action in actions or []:
            if isinstance(action, dict) and "items" in action:
                results.extend(action["items"])
            else:
                results.append(action)

        return results

    def _get_base_actions_list(self) -> list[UnfoldAction]:
        """
        Returns list of UnfoldAction objects for `actions_list`.
        """
        return [
            self.get_unfold_action(action)
            for action in self._extract_action_names(self.actions_list)
        ]

    def _get_base_actions_detail(self) -> list[UnfoldAction]:
        """
        Returns list of UnfoldAction objects for `actions_detail`.
        """
        return [
            self.get_unfold_action(action)
            for action in self._extract_action_names(self.actions_detail) or []
        ]

    def _get_base_actions_row(self) -> list[UnfoldAction]:
        """
        Returns list of UnfoldAction objects for `actions_row`.
        """
        return [
            self.get_unfold_action(action)
            for action in self._extract_action_names(self.actions_row) or []
        ]

    def _get_base_actions_submit_line(self) -> list[UnfoldAction]:
        """
        Returns list of UnfoldAction objects for `actions_submit_line`.
        """
        return [
            self.get_unfold_action(action)
            for action in self._extract_action_names(self.actions_submit_line) or []
        ]

    def _get_instance_method(self, method_name: str) -> Callable:
        """
        Searches for method on self instance based on method_name and returns it if it exists.
        If it does not exist or is not callable, it raises UnfoldException
        """
        try:
            method = getattr(self, method_name)
        except AttributeError as e:
            raise UnfoldException(
                f"Method {method_name} specified does not exist on current object"
            ) from e

        if not callable(method):
            raise UnfoldException(f"{method_name} is not callable")

        return method

    def _get_actions_navigation(
        self,
        provided_actions: list[str | dict],
        allowed_actions: list[UnfoldAction],
        object_id: str | None = None,
    ) -> list[str | dict]:
        """
        Builds navigation structure for the actions which is going to be provided to the template.
        """
        navigation = []

        def get_action_by_name(name: str) -> UnfoldAction:
            """
            Searches for an action in allowed_actions by its name.
            """
            for action in allowed_actions:
                full_action_name = (
                    f"{self.model._meta.app_label}_{self.model._meta.model_name}_{name}"
                )

                if action.action_name == full_action_name:
                    return action

        def get_action_path(action: UnfoldAction) -> str:
            """
            Returns the URL path for an action.
            """
            path_name = f"{self.admin_site.name}:{action.action_name}"

            if object_id:
                return reverse(path_name, args=(object_id,))

            return reverse(path_name)

        def get_action_attrs(action: UnfoldAction) -> dict:
            """
            Returns the attributes for an action which will be used in the template.
            """
            return {
                "title": action.description,
                "icon": action.icon,
                "variant": action.variant,
                "attrs": action.method.attrs,
                "path": get_action_path(action),
            }

        def build_dropdown(nav_item: dict) -> dict | None:
            """
            Builds a dropdown structure for the action.
            """
            dropdown = {
                "title": nav_item["title"],
                "icon": nav_item.get("icon"),
                "variant": nav_item.get("variant", ActionVariant.DEFAULT),
                "items": [],
            }

            for child in nav_item["items"]:
                if action := get_action_by_name(child):
                    dropdown["items"].append(get_action_attrs(action))

            if len(dropdown["items"]) > 0:
                return dropdown

        for nav_item in provided_actions:
            if isinstance(nav_item, str):
                if action := get_action_by_name(nav_item):
                    navigation.append(get_action_attrs(action))
            elif isinstance(nav_item, dict):
                if dropdown := build_dropdown(nav_item):
                    navigation.append(dropdown)

        return navigation

    def _filter_unfold_actions_by_permissions(
        self,
        request: HttpRequest,
        actions: list[UnfoldAction],
        object_id: int | str | None = None,
    ) -> list[UnfoldAction]:
        """
        Filters out actions that the user doesn't have access to.
        """
        filtered_actions = []

        for action in actions:
            if not hasattr(action.method, "allowed_permissions"):
                filtered_actions.append(action)
                continue

            permission_rules = []

            for permission in action.method.allowed_permissions:
                if "." in permission:
                    permission_rules.append(permission)
                else:
                    permission_rules.append(
                        getattr(self, f"has_{permission}_permission")
                    )

            permission_checks = []

            for permission_rule in permission_rules:
                if isinstance(permission_rule, str) and "." in permission_rule:
                    permission_checks.append(request.user.has_perm(permission_rule))
                elif object_id:
                    permission_checks.append(permission_rule(request, object_id))
                else:
                    permission_checks.append(permission_rule(request))

            if all(permission_checks):
                filtered_actions.append(action)

        return filtered_actions



UNFOLD_CONSTANCE_ADDITIONAL_FIELDS = {
    str: [
        "django.forms.CharField",
        {
            "widget": "unfold.widgets.UnfoldAdminTextInputWidget",
        },
    ],
    int: [
        "django.forms.IntegerField",
        {
            "widget": "unfold.widgets.UnfoldAdminIntegerFieldWidget",
        },
    ],
    bool: [
        "django.forms.BooleanField",
        {
            "widget": "unfold.widgets.UnfoldBooleanSwitchWidget",
            "required": False,
        },
    ],
    "file_field": [
        "django.forms.fields.FileField",
        {
            "widget": "unfold.widgets.UnfoldAdminFileFieldWidget",
        },
    ],
    "image_field": [
        "django.forms.fields.ImageField",
        {
            "widget": "unfold.widgets.UnfoldAdminImageFieldWidget",
        },
    ],
}



from typing import Any

from django.contrib import admin
from django.http import HttpRequest
from django.template.loader import render_to_string

from unfold.views import DatasetChangeList


class BaseDataset:
    tab = False

    def __init__(
        self, request: HttpRequest, extra_context: dict[str, Any] | None
    ) -> None:
        self.request = request
        self.extra_context = extra_context

        self.model_admin_instance = self.model_admin(
            model=self.model, admin_site=admin.site
        )
        self.model_admin_instance.extra_context = self.extra_context

    @property
    def contents(self) -> str:
        return render_to_string(
            "unfold/helpers/dataset.html",
            request=self.request,
            context={
                "dataset": self,
                "cl": self.cl(),
                "opts": self.model._meta,
            },
        )

    def cl(self) -> DatasetChangeList:
        list_display = self.model_admin_instance.get_list_display(self.request)
        list_display_links = self.model_admin_instance.get_list_display_links(
            self.request, list_display
        )
        sortable_by = self.model_admin_instance.get_sortable_by(self.request)
        search_fields = self.model_admin_instance.get_search_fields(self.request)
        cl = DatasetChangeList(
            request=self.request,
            model=self.model,
            model_admin=self.model_admin_instance,
            list_display=list_display,
            list_display_links=list_display_links,
            list_filter=[],
            date_hierarchy=[],
            search_fields=search_fields,
            list_select_related=[],
            list_per_page=10,
            list_max_show_all=False,
            list_editable=[],
            sortable_by=sortable_by,
            search_help_text=[],
        )
        cl.formset = None

        return cl

    @property
    def model_name(self) -> str:
        return self.model._meta.model_name

    @property
    def model_verbose_name(self) -> str:
        return self.model._meta.verbose_name_plural



from collections.abc import Generator
from typing import Any, Union

from django import forms
from django.contrib.admin.forms import (
    AdminAuthenticationForm,
)
from django.contrib.admin.forms import (
    AdminPasswordChangeForm as BaseAdminOwnPasswordChangeForm,
)
from django.contrib.admin.views.main import ChangeListSearchForm
from django.contrib.auth.forms import (
    AdminPasswordChangeForm as BaseAdminPasswordChangeForm,
)
from django.contrib.auth.models import User
from django.contrib.contenttypes.forms import BaseGenericInlineFormSet
from django.core.paginator import Page, Paginator
from django.db.models import QuerySet
from django.forms import BaseInlineFormSet
from django.http import HttpRequest

from unfold.fields import UnfoldAdminField, UnfoldAdminReadonlyField

try:
    from django.contrib.auth.forms import AdminUserCreationForm as BaseUserCreationForm
except ImportError:
    from django.contrib.auth.forms import UserCreationForm as BaseUserCreationForm
from django.contrib.admin.helpers import AdminForm as BaseAdminForm
from django.contrib.admin.helpers import Fieldline as BaseFieldline
from django.contrib.admin.helpers import Fieldset as BaseFieldset
from django.contrib.auth.forms import ReadOnlyPasswordHashWidget
from django.contrib.auth.forms import UserChangeForm as BaseUserChangeForm
from django.utils.safestring import mark_safe
from django.utils.translation import gettext_lazy as _

from unfold.widgets import (
    BASE_INPUT_CLASSES,
    INPUT_CLASSES,
    UnfoldAdminPasswordInput,
    UnfoldAdminRadioSelectWidget,
    UnfoldAdminSelectWidget,
)


class UnfoldReadOnlyPasswordHashWidget(ReadOnlyPasswordHashWidget):
    pass


class ActionForm(forms.Form):
    action = forms.ChoiceField(
        label="",
        widget=UnfoldAdminSelectWidget(
            {
                "class": " ".join(
                    [
                        "appearance-none",
                        "!bg-white/20",
                        "font-medium",
                        "grow",
                        "px-3",
                        "py-2",
                        "pr-8",
                        "rounded-default",
                        "!text-white",
                        "truncate",
                        "!outline-primary-400",
                        "dark:!outline-primary-700",
                        "*:text-base-700",
                        "lg:w-72",
                    ]
                ),
                "aria-label": _("Select action to run"),
                "x-model": "action",
            }
        ),
    )

    select_across = forms.BooleanField(
        label="",
        required=False,
        initial=0,
        widget=forms.HiddenInput({"class": "select-across"}),
    )


class AuthenticationForm(AdminAuthenticationForm):
    def __init__(
        self,
        request: HttpRequest | None = None,
        *args,
        **kwargs,
    ) -> None:
        super().__init__(request, *args, **kwargs)

        self.fields["username"].widget.attrs["class"] = " ".join(BASE_INPUT_CLASSES)
        self.fields["password"].widget.attrs["class"] = " ".join(BASE_INPUT_CLASSES)


class UserCreationForm(BaseUserCreationForm):
    def __init__(
        self,
        *args,
        **kwargs,
    ) -> None:
        super().__init__(*args, **kwargs)

        self.fields["password1"].widget = UnfoldAdminPasswordInput(
            attrs={"autocomplete": "new-password"}
        )
        self.fields["password2"].widget = UnfoldAdminPasswordInput(
            attrs={"autocomplete": "new-password"}
        )

        if "usable_password" in self.fields:
            self.fields["usable_password"].widget = UnfoldAdminRadioSelectWidget(
                choices=self.fields["usable_password"].choices,
            )


class UserChangeForm(BaseUserChangeForm):
    def __init__(
        self,
        *args,
        **kwargs,
    ) -> None:
        super().__init__(*args, **kwargs)
        self.fields["password"].widget = UnfoldReadOnlyPasswordHashWidget()

        self.fields["password"].help_text = _(
            "Raw passwords are not stored, so there is no way to see this "
            "user’s password, but you can change the password using "
            '<a href="{}" class="text-primary-600 dark:text-primary-500">this form</a>.'
        )

        password = self.fields.get("password")
        if password:
            password.help_text = mark_safe(password.help_text.format("../password/"))


class AdminPasswordChangeForm(BaseAdminPasswordChangeForm):
    def __init__(
        self,
        user: User,
        *args,
        **kwargs,
    ) -> None:
        super().__init__(user, *args, **kwargs)

        self.fields["password1"].widget.attrs["class"] = " ".join(INPUT_CLASSES)
        self.fields["password2"].widget.attrs["class"] = " ".join(INPUT_CLASSES)


class AdminOwnPasswordChangeForm(BaseAdminOwnPasswordChangeForm):
    def __init__(self, user: User, *args, **kwargs) -> None:
        super().__init__(user, *args, **kwargs)

        self.fields["old_password"].widget.attrs["class"] = " ".join(INPUT_CLASSES)
        self.fields["new_password1"].widget.attrs["class"] = " ".join(INPUT_CLASSES)
        self.fields["new_password2"].widget.attrs["class"] = " ".join(INPUT_CLASSES)


class AdminForm(BaseAdminForm):
    def __iter__(self) -> Generator["Fieldset", None, None]:
        for name, options in self.fieldsets:
            yield Fieldset(
                self.form,
                name,
                readonly_fields=self.readonly_fields,
                model_admin=self.model_admin,
                **options,
            )


class Fieldset(BaseFieldset):
    def __iter__(self) -> Generator["Fieldline", None, None]:
        for field in self.fields:
            yield Fieldline(
                self.form, field, self.readonly_fields, model_admin=self.model_admin
            )


class Fieldline(BaseFieldline):
    def __iter__(
        self,
    ) -> Generator[Union["UnfoldAdminReadonlyField", "UnfoldAdminField"], None, None]:
        for i, field in enumerate(self.fields):
            if field in self.readonly_fields:
                yield UnfoldAdminReadonlyField(
                    self.form, field, is_first=(i == 0), model_admin=self.model_admin
                )
            else:
                yield UnfoldAdminField(self.form, field, is_first=(i == 0))


class PaginationFormSetMixin:
    queryset: QuerySet | None = None
    request: HttpRequest | None = None
    per_page: int | None = None

    def __init__(
        self,
        request: HttpRequest | None = None,
        per_page: int | None = None,
        *args,
        **kwargs,
    ):
        self.request = request
        self.per_page = per_page

        super().__init__(*args, **kwargs)

        if self.per_page:
            self.paginator = Paginator(self.queryset, self.per_page)
            self.page = self.get_page(self.paginator, self.get_page_num())
            self._queryset = self.page.object_list

    def get_pagination_key(self) -> str:
        return f"{self.prefix}-page"

    def get_page_num(self) -> int:
        page = self.request.GET.get(self.get_pagination_key())
        if page and page.isnumeric() and page > "0":
            return int(page)

        page = self.request.POST.get(self.get_pagination_key())
        if page and page.isnumeric() and page > "0":
            return int(page)

        return 1

    def get_page(self, paginator: Paginator, page: int) -> Page:
        if page <= paginator.num_pages:
            return paginator.page(page)

        return paginator.page(1)


class PaginationInlineFormSet(PaginationFormSetMixin, BaseInlineFormSet):
    pass


class PaginationGenericInlineFormSet(PaginationFormSetMixin, BaseGenericInlineFormSet):
    pass


class DatasetChangeListSearchForm(ChangeListSearchForm):
    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)

        from django.contrib.admin.views.main import SEARCH_VAR

        self.fields = {
            SEARCH_VAR: forms.CharField(required=False, strip=False),
        }



import copy
import time
from collections.abc import Callable
from http import HTTPStatus
from typing import Any
from urllib.parse import parse_qs, urlparse

from django.contrib.admin import AdminSite
from django.core.cache import cache
from django.core.paginator import Paginator
from django.core.validators import EMPTY_VALUES
from django.http import HttpRequest, HttpResponse
from django.template.response import TemplateResponse
from django.urls import URLPattern, path, reverse, reverse_lazy
from django.utils.functional import lazy
from django.utils.module_loading import import_string
from django.utils.text import slugify

from unfold.dataclasses import DropdownItem, Favicon, SearchResult

try:
    from django.contrib.auth.decorators import login_not_required
except ImportError:

    def login_not_required(func: Callable) -> Callable:
        return func


from unfold.settings import get_config
from unfold.utils import convert_color
from unfold.widgets import (
    BUTTON_CLASSES,
    CHECKBOX_CLASSES,
    FILE_CLASSES,
    INPUT_CLASSES,
    RADIO_CLASSES,
    SWITCH_CLASSES,
)


class UnfoldAdminSite(AdminSite):
    default_site = "unfold.admin.UnfoldAdminSite"
    settings_name = "UNFOLD"

    def __init__(self, name: str = "admin") -> None:
        from unfold.forms import AuthenticationForm

        super().__init__(name)

        if self.login_form is None:
            self.login_form = AuthenticationForm

    def get_urls(self) -> list[URLPattern]:
        extra_urls = []

        if hasattr(self, "extra_urls") and callable(self.extra_urls):
            extra_urls = self.extra_urls()

        urlpatterns = (
            [
                path("search/", self.admin_view(self.search), name="search"),
                path(
                    "toggle-sidebar/",
                    self.admin_view(self.toggle_sidebar),
                    name="toggle_sidebar",
                ),
            ]
            + extra_urls
            + super().get_urls()
        )

        return urlpatterns

    def each_context(self, request: HttpRequest) -> dict[str, Any]:
        context = super().each_context(request)

        sidebar_config = self._get_config("SIDEBAR", request)
        data = {
            "form_classes": {
                "text_input": " ".join(INPUT_CLASSES),
                "checkbox": " ".join(CHECKBOX_CLASSES),
                "button": " ".join(BUTTON_CLASSES),
                "radio": " ".join(RADIO_CLASSES),
                "switch": " ".join(SWITCH_CLASSES),
                "file": " ".join(FILE_CLASSES),
            },
            "site_title": self._get_config("SITE_TITLE", request),
            "site_header": self._get_config("SITE_HEADER", request),
            "site_url": self._get_config("SITE_URL", request),
            "site_subheader": self._get_config("SITE_SUBHEADER", request),
            "site_dropdown": self._get_site_dropdown_items("SITE_DROPDOWN", request),
            "site_logo": self._get_theme_images("SITE_LOGO", request),
            "site_icon": self._get_theme_images("SITE_ICON", request),
            "site_symbol": self._get_config("SITE_SYMBOL", request),
            "site_favicons": self._get_favicons("SITE_FAVICONS", request),
            "login_image": self._get_value(
                get_config(self.settings_name)["LOGIN"].get("image"), request
            ),
            "show_history": self._get_config("SHOW_HISTORY", request),
            "show_view_on_site": self._get_config("SHOW_VIEW_ON_SITE", request),
            "show_languages": self._get_config("SHOW_LANGUAGES", request),
            "show_back_button": self._get_config("SHOW_BACK_BUTTON", request),
            "theme": self._get_config("THEME", request),
            "border_radius": self._get_config("BORDER_RADIUS", request),
            "colors": self._get_colors("COLORS", request),
            "environment": self._get_config("ENVIRONMENT", request),
            "environment_title_prefix": self._get_config(
                "ENVIRONMENT_TITLE_PREFIX", request
            ),
            "languages_list": self._get_value(
                self._get_config("LANGUAGES", request).get("navigation"), request
            ),
            "languages_action": self._get_value(
                self._get_config("LANGUAGES", request).get("action"), request
            ),
            "account_links": self._get_account_links(request),
            "tab_list": self.get_tabs_list(request),
            "styles": self._get_list("STYLES", request),
            "scripts": self._get_list("SCRIPTS", request),
            "command_show_history": self._get_config("COMMAND", request).get(
                "show_history"
            ),
            "sidebar_command_search": self._get_config("SIDEBAR", request).get(
                "command_search"
            ),
            "sidebar_show_all_applications": self._get_value(
                sidebar_config.get("show_all_applications"), request
            ),
            "sidebar_show_search": self._get_value(
                sidebar_config.get("show_search"), request
            ),
            "sidebar_navigation": self.get_sidebar_list(request)
            if self.has_permission(request)
            else [],
        }

        context.update(data)

        if hasattr(self, "extra_context") and callable(self.extra_context):
            return self.extra_context(context, request)

        return context

    def index(
        self, request: HttpRequest, extra_context: dict[str, Any] | None = None
    ) -> TemplateResponse:
        app_list = self.get_app_list(request)

        context = {
            **self.each_context(request),
            "title": self.index_title,
            "subtitle": None,
            "app_list": app_list,
            "index": True,
            **(extra_context or {}),
        }

        dashboard_callback = get_config(self.settings_name)["DASHBOARD_CALLBACK"]

        if isinstance(dashboard_callback, str):
            context = import_string(dashboard_callback)(request, context)

        request.current_app = self.name

        return TemplateResponse(
            request, self.index_template or "admin/index.html", context
        )

    def toggle_sidebar(
        self, request: HttpRequest, extra_context: dict[str, Any] | None = None
    ) -> HttpResponse:
        if "toggle_sidebar" not in request.session:
            request.session["toggle_sidebar"] = True
        else:
            request.session["toggle_sidebar"] = not request.session["toggle_sidebar"]

        return HttpResponse(status=HTTPStatus.OK)

    def _search_apps(
        self, app_list: list[dict[str, Any]], search_term: str
    ) -> list[SearchResult]:
        results = []
        apps = []

        for app in app_list:
            if search_term in app["name"].lower():
                apps.append(app)
                continue

            models = []

            for model in app["models"]:
                if search_term in model["name"].lower():
                    models.append(model)

            if len(models) > 0:
                app["models"] = models
                apps.append(app)

        for app in apps:
            for model in app["models"]:
                results.append(
                    SearchResult(
                        title=str(model["name"]),
                        description=app["name"],
                        link=model["admin_url"],
                        icon="tag",
                    )
                )

        return results

    def _search_models(
        self,
        request: HttpRequest,
        app_list: list[dict[str, Any]],
        search_term: str,
        allowed_models: list[str] | None = None,
    ) -> list[SearchResult]:
        results = []

        for app in app_list:
            for model in app["models"]:
                # Skip models which are not allowed
                if isinstance(allowed_models, list | tuple):
                    if model["model"]._meta.label.lower() not in [
                        m.lower() for m in allowed_models
                    ]:
                        continue

                admin_instance = self._registry.get(model["model"])
                search_fields = admin_instance.get_search_fields(request)

                if not search_fields:
                    continue

                pks = []

                qs = admin_instance.get_queryset(request)
                search_results, _has_duplicates = admin_instance.get_search_results(
                    request, qs, search_term
                )

                for item in search_results:
                    if item.pk in pks:
                        continue

                    pks.append(item.pk)

                    link = reverse_lazy(
                        f"{self.name}:{admin_instance.model._meta.app_label}_{admin_instance.model._meta.model_name}_change",
                        args=(item.pk,),
                    )

                    results.append(
                        SearchResult(
                            title=str(item),
                            description=f"{item._meta.app_label.capitalize()} - {item._meta.verbose_name.capitalize()}",
                            link=link,
                            icon="data_object",
                        )
                    )

        return results

    def search(
        self, request: HttpRequest, extra_context: dict[str, Any] | None = None
    ) -> TemplateResponse:
        start_time = time.time()

        CACHE_TIMEOUT = 5 * 60
        PER_PAGE = 100

        search_term = request.GET.get("s")
        extended_search = "extended" in request.GET
        app_list = super().get_app_list(request)
        template_name = "unfold/helpers/search_results.html"

        if search_term in EMPTY_VALUES:
            return HttpResponse()

        search_term = search_term.lower()
        cache_key = f"unfold_search_{request.user.pk}_{slugify(search_term)}"
        cache_results = cache.get(cache_key)

        if extended_search:
            template_name = "unfold/helpers/command_results.html"

        if cache_results:
            results = cache_results
        else:
            results = self._search_apps(app_list, search_term)

            if extended_search:
                if search_callback := self._get_config("COMMAND", request).get(
                    "search_callback"
                ):
                    results.extend(
                        self._get_value(search_callback, request, search_term)
                    )

                search_models = self._get_value(
                    self._get_config("COMMAND", request).get("search_models"), request
                )

                if search_models is True or isinstance(search_models, list | tuple):
                    allowed_models = (
                        search_models
                        if isinstance(search_models, list | tuple)
                        else None
                    )

                    results.extend(
                        self._search_models(
                            request, app_list, search_term, allowed_models
                        )
                    )

            cache.set(cache_key, results, timeout=CACHE_TIMEOUT)

        execution_time = time.time() - start_time
        paginator = Paginator(results, PER_PAGE)

        show_history = self._get_value(
            self._get_config("COMMAND", request).get("show_history"), request
        )

        return TemplateResponse(
            request,
            template=template_name,
            context={
                "page_obj": paginator,
                "results": paginator.page(request.GET.get("page", 1)),
                "page_counter": (int(request.GET.get("page", 1)) - 1) * PER_PAGE,
                "execution_time": execution_time,
                "command_show_history": show_history,
            },
            headers={
                "HX-Trigger": "search",
            },
        )

    def password_change(
        self, request: HttpRequest, extra_context: dict[str, Any] | None = None
    ) -> HttpResponse:
        from django.contrib.auth.views import PasswordChangeView

        from .forms import AdminOwnPasswordChangeForm

        url = reverse(f"{self.name}:password_change_done", current_app=self.name)
        defaults = {
            "form_class": AdminOwnPasswordChangeForm,
            "success_url": url,
            "extra_context": {**self.each_context(request), **(extra_context or {})},
        }
        if self.password_change_template is not None:
            defaults["template_name"] = self.password_change_template
        request.current_app = self.name
        return PasswordChangeView.as_view(**defaults)(request)

    def get_sidebar_list(self, request: HttpRequest) -> list[dict[str, Any]]:
        navigation = self._get_value(
            self._get_config("SIDEBAR", request).get("navigation"), request
        )
        tabs = self._get_value(self._get_config("TABS", request), request) or []
        results = []

        for group in copy.deepcopy(navigation):
            group["items"] = self._get_navigation_items(request, group["items"], tabs)
            results.append(group)

        return results

    def _get_navigation_items(
        self, request: HttpRequest, items: list[dict], tabs: list[dict] = None
    ) -> list:
        allowed_items = []

        for item in items:
            link = item.get("link")

            if "active" in item:
                item["active"] = self._get_value(item["active"], request)
            else:
                item["active"] = self._get_is_active(
                    request, item.get("link_callback") or link
                )

            # Checks if any tab item is active and then marks the sidebar link as active
            if tabs and self._get_is_tab_active(request, tabs, link):
                item["active"] = True

            # Link callback
            if isinstance(link, Callable):
                item["link_callback"] = lazy(link)(request)

            # Permission callback
            item["has_permission"] = self._call_permission_callback(
                item.get("permission"), request
            )

            # Badge callbacks
            if "badge" in item and isinstance(item["badge"], str):
                try:
                    callback = import_string(item["badge"])
                    item["badge_callback"] = lazy(callback)(request)
                except ImportError:
                    pass

            # Process nested items
            if "items" in item:
                item["items"] = self._get_navigation_items(request, item["items"])

            allowed_items.append(item)

        return allowed_items

    def _get_account_links(self, request: HttpRequest) -> list[dict[str, Any]]:
        links = []

        navigation = self._get_value(
            get_config(self.settings_name)["ACCOUNT"].get("navigation"), request
        )

        for item in navigation:
            links.append(
                {
                    "title": self._get_value(item["title"], request),
                    "link": self._get_value(item["link"], request),
                }
            )

        return links

    def get_tabs_list(self, request: HttpRequest) -> list[dict[str, Any]]:
        tabs = copy.deepcopy(self._get_config("TABS", request))

        if not tabs:
            return []

        for tab in tabs:
            allowed_items = []

            for item in tab["items"]:
                item["has_permission"] = self._call_permission_callback(
                    item.get("permission"), request
                )

                if isinstance(item["link"], Callable):
                    item["link_callback"] = lazy(item["link"])(request)

                if "active" not in item:
                    item["active"] = self._get_is_active(
                        request, item.get("link_callback") or item["link"], True
                    )
                else:
                    item["active"] = self._get_value(item["active"], request)

                allowed_items.append(item)

            tab["items"] = allowed_items

        return tabs

    def _call_permission_callback(
        self, callback: str | Callable | None, request: HttpRequest
    ) -> bool:
        if callback is None:
            return True

        if isinstance(callback, str):
            try:
                callback = import_string(callback)
            except ImportError:
                pass

        if isinstance(callback, str) or isinstance(callback, Callable):
            # We are not able to use here "is" because type is lazy loaded function
            if lazy(callback)(request) == True:  # noqa: E712
                return True

        return False

    def _replace_values(self, target: dict, source: dict, request: HttpRequest):
        for key in source.keys():
            if source[key] is not None and callable(source[key]):
                target[key] = source[key](request)
            else:
                target[key] = source[key]

        return target

    def _get_is_active(
        self, request: HttpRequest, link: str | Callable, is_tab: bool = False
    ) -> bool:
        if not isinstance(link, str):
            link = str(link)

        index_path = reverse_lazy(f"{self.name}:index")
        link_path = urlparse(link).path

        # Dashboard
        if link_path == request.path == index_path:
            return True

        if link_path != "" and link_path in request.path and link_path != index_path:
            query_params = parse_qs(urlparse(link).query)
            request_params = parse_qs(request.GET.urlencode())

            # In case of tabs, we need to check if the query params are the same
            if is_tab and not all(
                request_params.get(k) == v for k, v in query_params.items()
            ):
                return False

            return True

        return False

    def _get_is_tab_active(
        self, request: HttpRequest, tabs: list[dict], link: str
    ) -> bool:
        for tab in tabs:
            has_primary_link = False
            has_tab_link_active = False

            for tab_item in tab["items"]:
                if link == tab_item["link"]:
                    has_primary_link = True
                    continue

                if self._get_is_active(
                    request, tab_item.get("link_callback") or tab_item["link"]
                ):
                    has_tab_link_active = True
                    continue

            if has_primary_link and has_tab_link_active:
                return True

        return False

    def _get_config(self, key: str, *args) -> Any:
        config = get_config(self.settings_name)

        if key in config and config[key]:
            return self._get_value(config[key], *args)

    def _get_theme_images(self, key: str, *args: Any) -> dict[str, str] | str | None:
        images = self._get_config(key, *args)

        if isinstance(images, dict):
            if "light" in images and "dark" in images:
                return {
                    "light": self._get_value(images["light"], *args),
                    "dark": self._get_value(images["dark"], *args),
                }

            return None

        return images

    def _get_colors(self, key: str, *args) -> dict[str, dict[str, str]]:
        colors = self._get_config(key, *args)

        for name, weights in colors.items():
            weights = self._get_value(weights, *args)
            colors[name] = weights

            for weight, value in weights.items():
                colors[name][weight] = convert_color(value)

        return colors

    def _get_list(self, key: str, *args) -> list[Any]:
        items = get_config(self.settings_name)[key]

        if isinstance(items, list):
            return [self._get_value(item, *args) for item in items]

        return []

    def _get_favicons(self, key: str, *args) -> list[Favicon]:
        favicons = self._get_config(key, *args)

        if not favicons:
            return []

        return [
            Favicon(
                href=self._get_value(item["href"], *args),
                rel=item.get("rel"),
                sizes=item.get("sizes"),
                type=item.get("type"),
            )
            for item in favicons
        ]

    def _get_site_dropdown_items(self, key: str, *args) -> list[dict[str, Any]]:
        items = self._get_config(key, *args)

        if not items:
            return []

        return [
            DropdownItem(
                title=item.get("title"),
                link=self._get_value(item["link"], *args),
                icon=item.get("icon"),
                attrs=item.get("attrs"),
            )
            for item in items
        ]

    def _get_value(self, value: str | Callable | None, *args: Any) -> str | None:
        if value is None:
            return None

        if isinstance(value, str):
            try:
                callback = import_string(value)
                return callback(*args)
            except ImportError:
                pass

            return value

        if isinstance(value, Callable):
            return value(*args)

        return value



import datetime
import decimal
import json
from collections.abc import Iterable
from typing import Any

from django.conf import settings
from django.db import models
from django.db.models import Model
from django.template.loader import render_to_string
from django.utils import formats, timezone
from django.utils.hashable import make_hashable
from django.utils.html import format_html
from django.utils.safestring import SafeText, mark_safe

from .exceptions import UnfoldException

try:
    from djmoney.models.fields import MoneyField
    from djmoney.money import Money
except ImportError:
    MoneyField = None
    Money = None


def _boolean_icon(field_val: Any) -> str:
    return render_to_string("unfold/helpers/boolean.html", {"value": field_val})


def display_for_header(value: Iterable, empty_value_display: str) -> SafeText:
    if not isinstance(value, list) and not isinstance(value, tuple):
        raise UnfoldException("Display header requires list or tuple")

    return mark_safe(
        render_to_string(
            "unfold/helpers/display_header.html",
            {
                "value": value,
            },
        )
    )


def display_for_dropdown(
    result: Model, field_name: str, value: Iterable, empty_value_display: str
) -> SafeText:
    return render_to_string(
        "unfold/helpers/display_dropdown.html",
        {
            "instance": result,
            "field_name": field_name,
            "value": value,
        },
    )


def display_for_label(value: Any, empty_value_display: str, label: Any) -> SafeText:
    label_type = None
    multiple = False

    if isinstance(label, dict):
        if isinstance(value, tuple):
            try:
                label_type = label[value[0]]
                value = value[1]
            except KeyError:
                value = value[0]
        elif value in label:
            label_type = label[value]

    if isinstance(value, tuple) or isinstance(value, list):
        multiple = True

    return mark_safe(
        render_to_string(
            "unfold/helpers/display_label.html",
            {
                "label": value,
                "label_type": label_type,
                "multiple": multiple,
            },
        )
    )


def display_for_value(
    value: Any, empty_value_display: str, boolean: bool = False
) -> str:
    if boolean:
        return _boolean_icon(value)
    elif value is None:
        return empty_value_display
    elif isinstance(value, bool):
        return str(value)
    elif isinstance(value, datetime.datetime):
        return formats.localize(timezone.template_localtime(value))
    elif isinstance(value, datetime.date | datetime.time):
        return formats.localize(value)
    elif Money is not None and isinstance(value, Money):
        return str(value)
    elif isinstance(value, int | decimal.Decimal | float):
        return formats.number_format(value)
    elif isinstance(value, list | tuple):
        return ", ".join(str(v) for v in value)
    else:
        return str(value)


def display_for_field(value: Any, field: Any, empty_value_display: str) -> str:
    if getattr(field, "flatchoices", None):
        try:
            return dict(field.flatchoices).get(value, empty_value_display)
        except TypeError:
            # Allow list-like choices.
            flatchoices = make_hashable(field.flatchoices)
            value = make_hashable(value)
            return dict(flatchoices).get(value, empty_value_display)
    elif isinstance(field, models.BooleanField):
        return _boolean_icon(value)
    elif value is None or value == "":
        return empty_value_display
    elif isinstance(field, models.DateTimeField):
        return formats.localize(timezone.template_localtime(value))
    elif isinstance(field, models.DateField | models.TimeField):
        return formats.localize(value)
    elif MoneyField is not None and isinstance(field, MoneyField):
        return str(value)
    elif isinstance(field, models.DecimalField):
        return formats.number_format(value, field.decimal_places)
    elif isinstance(field, models.IntegerField | models.FloatField):
        return formats.number_format(value)
    elif isinstance(field, models.FileField) and value:
        return format_html('<a href="{}">{}</a>', value.url, value)
    elif isinstance(field, models.JSONField) and value:
        try:
            return json.dumps(value, ensure_ascii=False, cls=field.encoder)
        except TypeError:
            return display_for_value(value, empty_value_display)
    else:
        return display_for_value(value, empty_value_display)


def prettify_json(data: Any, encoder: Any) -> str | None:
    try:
        from pygments import highlight
        from pygments.formatters import HtmlFormatter
        from pygments.lexers import JsonLexer
    except ImportError:
        return None

    def format_response(response: str, theme: str) -> str:
        formatter = HtmlFormatter(
            style=theme,
            noclasses=True,
            nobackground=True,
            prestyles="white-space: pre-wrap; word-wrap: break-word;",
        )
        return highlight(response, JsonLexer(), formatter)

    response = json.dumps(data, sort_keys=True, indent=4, cls=encoder)

    return mark_safe(
        f'<div class="block dark:hidden">{format_response(response, "colorful")}</div>'
        f'<div class="hidden dark:block">{format_response(response, "monokai")}</div>'
    )


def parse_date_str(value: str) -> datetime.date | None:
    for format in settings.DATE_INPUT_FORMATS:
        try:
            return datetime.datetime.strptime(value, format).date()
        except (ValueError, TypeError):
            continue


def parse_datetime_str(value: str) -> datetime.datetime | None:
    for format in settings.DATETIME_INPUT_FORMATS:
        try:
            return datetime.datetime.strptime(value, format)
        except (ValueError, TypeError):
            continue


def hex_to_rgb(hex_color: str) -> list[int]:
    hex_color = hex_color.lstrip("#")

    r = int(hex_color[0:2], 16)
    g = int(hex_color[2:4], 16)
    b = int(hex_color[4:6], 16)

    return (r, g, b)


def hex_to_values(value: str) -> str:
    return ", ".join(str(item) for item in hex_to_rgb(value))


def convert_color(value: str) -> str:
    if value[0] == "#":
        return f"rgb({hex_to_values(value)})"
    elif value.startswith("rgb") or value.startswith("oklch"):
        return value
    elif isinstance(value, str) and all(part.isdigit() for part in value.split()):
        return f"rgb({', '.join(value.split(' '))})"
    elif isinstance(value, str) and all(
        part.strip().isdigit() for part in value.split(",")
    ):
        return f"rgb({', '.join(part.strip() for part in value.split(','))})"

    return value



from django.contrib.admin.utils import label_for_field, lookup_field
from django.db.models import Model
from django.http import HttpRequest
from django.template.loader import render_to_string

from unfold.utils import display_for_field


class BaseSection:
    def __init__(self, request: HttpRequest, instance: Model) -> None:
        self.request = request
        self.instance = instance

    def render(self) -> str:
        raise NotImplementedError(
            "Section subclasses must implement the render method."
        )


class TableSection(BaseSection):
    fields = []
    related_name = None
    verbose_name = None
    height = None

    def render(self) -> str:
        results = getattr(self.instance, self.related_name)
        headers = []
        rows = []

        for field_name in self.fields:
            if hasattr(self, field_name):
                if hasattr(getattr(self, field_name), "short_description"):
                    headers.append(getattr(self, field_name).short_description)
                else:
                    headers.append(field_name)
            else:
                headers.append(label_for_field(field_name, results.model))

        for result in results.all():
            row = []

            for field_name in self.fields:
                if hasattr(self, field_name):
                    row.append(getattr(self, field_name)(result))
                else:
                    field, attr, value = lookup_field(field_name, result)
                    row.append(display_for_field(value, field, "-"))

            rows.append(row)

        context = {
            "request": self.request,
            "table": {
                "headers": headers,
                "rows": rows,
            },
        }

        if hasattr(self, "verbose_name") and self.verbose_name:
            context["title"] = self.verbose_name

        if hasattr(self, "height") and self.height:
            context["height"] = self.height

        return render_to_string(
            "unfold/components/table.html",
            context=context,
        )


class TemplateSection(BaseSection):
    template_name = None

    def render(self) -> str:
        return render_to_string(
            self.template_name,
            context={
                "request": self.request,
                "instance": self.instance,
            },
        )



from collections.abc import Iterable
from typing import Any, Protocol


class ActionFunction(Protocol):
    """
    Type that specifies functions that are used as actions (annotated with @actions decorator)
    These are functions (__call__) and have 4 additional attributes
    (allowed_permissions, short_description, url_path and attrs)
    """

    allowed_permissions: Iterable[str]
    short_description: str
    url_path: str
    attrs: dict[str, Any]
    icon: str | None = None

    def __call__(self, *args, **kwargs):
        pass


FieldsetsType = (
    list[tuple[str | None, dict[str, Any]]] | tuple[tuple[str | None, dict[str, Any]]]
)



from collections.abc import Callable
from typing import Any

from django.contrib.admin.options import VERTICAL
from django.contrib.admin.sites import AdminSite
from django.contrib.admin.widgets import (
    AdminBigIntegerFieldWidget,
    AdminDateWidget,
    AdminEmailInputWidget,
    AdminFileWidget,
    AdminIntegerFieldWidget,
    AdminRadioSelect,
    AdminSplitDateTime,
    AdminTextareaWidget,
    AdminTextInputWidget,
    AdminTimeWidget,
    AdminURLFieldWidget,
    AdminUUIDInputWidget,
    ForeignKeyRawIdWidget,
    RelatedFieldWidgetWrapper,
)
from django.db.models.fields.reverse_related import ForeignObjectRel
from django.forms import (
    CheckboxInput,
    CheckboxSelectMultiple,
    MultiWidget,
    NullBooleanSelect,
    NumberInput,
    PasswordInput,
    Select,
    SelectMultiple,
)
from django.utils.translation import gettext_lazy as _

from .exceptions import UnfoldException

BUTTON_CLASSES = [
    "border",
    "cursor-pointer",
    "font-medium",
    "px-3",
    "py-2",
    "rounded-default",
    "text-center",
    "whitespace-nowrap",
    "bg-primary-600",
    "border-transparent",
    "text-white",
]

LABEL_CLASSES = [
    "block",
    "font-semibold",
    "mb-2",
    "text-font-important-light",
    "text-sm",
    "dark:text-font-important-dark",
]

CHECKBOX_LABEL_CLASSES = [
    "font-semibold",
    "ml-2",
    "text-sm",
    "text-font-important-light",
    "dark:text-font-important-dark",
]

BASE_CLASSES = [
    "border",
    "border-base-200",
    "bg-white",
    "font-medium",
    "min-w-20",
    "placeholder-base-400",
    "rounded-default",
    "shadow-xs",
    "text-font-default-light",
    "text-sm",
    "focus:outline-2",
    "focus:-outline-offset-2",
    "focus:outline-primary-600",
    "group-[.errors]:border-red-600",
    "focus:group-[.errors]:outline-red-600",
    "dark:bg-base-900",
    "dark:border-base-700",
    "dark:text-font-default-dark",
    "dark:group-[.errors]:border-red-500",
    "dark:focus:group-[.errors]:outline-red-500",
    "dark:scheme-dark",
    "group-[.primary]:border-transparent",
    "disabled:!bg-base-50",
    "dark:disabled:!bg-base-800",
]

BASE_INPUT_CLASSES = [
    *BASE_CLASSES,
    "px-3",
    "py-2",
    "w-full",
]

INPUT_CLASSES = [*BASE_INPUT_CLASSES, "max-w-2xl"]

DATETIME_CLASSES = [*BASE_INPUT_CLASSES, "min-w-52"]

COLOR_CLASSES = [*BASE_CLASSES, "h-[38px]", "px-2", "py-2", "w-32"]

INPUT_CLASSES_READONLY = [*BASE_INPUT_CLASSES, "bg-base-50"]

TEXTAREA_CLASSES = [
    *BASE_INPUT_CLASSES,
    "max-w-4xl",
    "appearance-none",
    "expandable",
    "transition",
    "transition-height",
    "duration-75",
    "ease-in-out",
]

TEXTAREA_EXPANDABLE_CLASSES = [
    "block",
    "field-sizing-content",
    "min-h-[38px]",
]

SELECT_CLASSES = [
    *BASE_INPUT_CLASSES,
    "pr-8!",
    "max-w-2xl",
    "appearance-none",
    "truncate",
]

PROSE_CLASSES = [
    "font-normal",
    "whitespace-normal",
    "prose-sm",
    "prose-blockquote:border-l-4",
    "prose-blockquote:not-italic",
    "prose-pre:bg-base-50",
    "prose-pre:rounded-default",
    "prose-headings:font-medium",
    "prose-a:text-primary-600",
    "prose-headings:font-medium",
    "prose-headings:text-base-700",
    "prose-ol:list-decimal",
    "prose-ul:list-disc",
    "prose-strong:text-base-700",
    "dark:prose-pre:bg-base-800",
    "dark:prose-blockquote:border-base-700",
    "dark:prose-blockquote:text-base-300",
    "dark:prose-headings:text-base-200",
    "dark:prose-strong:text-base-200",
]

CHECKBOX_CLASSES = [
    "appearance-none",
    "bg-white",
    "block",
    "border",
    "border-base-300",
    "cursor-pointer",
    "h-4",
    "min-w-4",
    "relative",
    "rounded-[4px]",
    "shadow-xs",
    "w-4",
    "hover:border-base-400",
    "dark:bg-base-700",
    "dark:border-base-500",
    "dark:checked:after:text-white",
    "focus:outline",
    "focus:outline-2",
    "focus:outline-offset-2",
    "focus:outline-primary-500",
    "after:absolute",
    "after:content-['done']",
    "after:flex!",
    "after:h-4",
    "after:items-center",
    "after:justify-center",
    "after:leading-none",
    "after:material-symbols-outlined",
    "after:-ml-px",
    "after:-mt-px",
    "after:text-sm!",
    "after:text-white",
    "after:transition-all",
    "after:w-4",
    "dark:after:text-base-700",
    "checked:bg-primary-600",
    "dark:checked:bg-primary-600",
    "checked:border-primary-600",
    "dark:checked:border-primary-600",
    "checked:transition-all",
    "checked:hover:border-primary-600",
]

RADIO_CLASSES = [
    "appearance-none",
    "bg-white",
    "block",
    "border",
    "border-base-300",
    "cursor-pointer",
    "h-4",
    "min-w-4",
    "relative",
    "rounded-full",
    "w-4",
    "dark:bg-base-700",
    "dark:border-base-500",
    "hover:border-base-400",
    "focus:outline",
    "focus:outline-2",
    "focus:outline-offset-2",
    "focus:outline-primary-500",
    "after:absolute",
    "after:bg-transparent",
    "after:content-['']",
    "after:flex",
    "after:h-2",
    "after:items-center",
    "after:justify-center",
    "after:leading-none",
    "after:left-1/2",
    "after:rounded-full",
    "after:text-white",
    "after:top-1/2",
    "after:transition-all",
    "after:-translate-x-1/2",
    "after:-translate-y-1/2",
    "after:text-sm",
    "after:w-2",
    "dark:after:text-base-700",
    "dark:after:bg-transparent",
    "checked:bg-primary-600",
    "checked:border-primary-600",
    "checked:transition-all",
    "checked:after:bg-white",
    "dark:checked:after:bg-base-900",
    "checked:hover:border-base-900/20",
]

SWITCH_CLASSES = [
    "appearance-none",
    "bg-base-300",
    "block",
    "cursor-pointer",
    "h-5",
    "relative",
    "rounded-full",
    "transition-colors",
    "w-8",
    "min-w-8",
    "disabled:cursor-not-allowed",
    "disabled:opacity-50",
    "focus:outline-none",
    "after:absolute",
    "after:bg-white",
    "after:content-['']",
    "after:bg-red-300",
    "after:h-3",
    "after:rounded-full",
    "after:shadow-xs",
    "after:transition-all",
    "after:left-1",
    "after:top-1",
    "after:w-3",
    "checked:bg-green-500",
    "checked:after:left-4",
    "dark:bg-base-600",
    "dark:checked:bg-green-700",
]

FILE_CLASSES = [
    "border",
    "border-base-200",
    "flex",
    "grow",
    "items-center",
    "overflow-hidden",
    "rounded-default",
    "shadow-xs",
    "max-w-2xl",
    "focus-within:outline-2",
    "focus-within:-outline-offset-2",
    "focus-within:outline-primary-600",
    "group-[.errors]:border-red-600",
    "focus-within:group-[.errors]:outline-red-500",
    "dark:border-base-700",
    "dark:group-[.errors]:border-red-500",
    "dark:focus-within:group-[.errors]:outline-red-500",
]


class UnfoldPrefixSuffixMixin:
    def get_context(self, name, value, attrs):
        context = super().get_context(name, value, attrs)
        widget = context["widget"]

        if "prefix" in self.attrs:
            widget["prefix"] = self.attrs["prefix"]
            del self.attrs["prefix"]

        if "prefix_icon" in self.attrs:
            widget["prefix_icon"] = self.attrs["prefix_icon"]
            self.attrs["class"] = " ".join([self.attrs["class"], "pl-9"])
            del self.attrs["prefix_icon"]

        if "suffix" in self.attrs:
            widget["suffix"] = self.attrs["suffix"]
            del self.attrs["suffix"]

        if "suffix_icon" in self.attrs:
            widget["suffix_icon"] = self.attrs["suffix_icon"]
            self.attrs["class"] = " ".join([self.attrs["class"], "pr-9"])
            del self.attrs["suffix_icon"]

        widget.update(
            {
                "name": name,
                "is_hidden": self.is_hidden,
                "required": self.is_required,
                "value": self.format_value(value),
                "attrs": self.build_attrs(self.attrs, attrs),
                "template_name": self.template_name,
            }
        )

        return {
            "widget": widget,
        }


class UnfoldAdminTextInputWidget(UnfoldPrefixSuffixMixin, AdminTextInputWidget):
    template_name = "unfold/widgets/text.html"

    def __init__(self, attrs: dict[str, Any] | None = None) -> None:
        super().__init__(
            attrs={
                **(attrs or {}),
                "class": " ".join(
                    [*INPUT_CLASSES, attrs.get("class", "") if attrs else ""]
                ),
            }
        )


class UnfoldAdminURLInputWidget(AdminURLFieldWidget):
    template_name = "unfold/widgets/url.html"

    def __init__(self, attrs: dict[str, Any] | None = None) -> None:
        super().__init__(
            attrs={
                **(attrs or {}),
                "class": " ".join(
                    [*INPUT_CLASSES, attrs.get("class", "") if attrs else ""]
                ),
            }
        )


class UnfoldAdminColorInputWidget(AdminTextInputWidget):
    def __init__(self, attrs: dict[str, Any] | None = None) -> None:
        super().__init__(
            attrs={
                **(attrs or {}),
                "type": "color",
                "class": " ".join(
                    [*COLOR_CLASSES, attrs.get("class", "") if attrs else ""]
                ),
            }
        )


class UnfoldAdminUUIDInputWidget(AdminUUIDInputWidget):
    def __init__(self, attrs: dict[str, Any] | None = None) -> None:
        super().__init__(
            attrs={
                **(attrs or {}),
                "class": " ".join(
                    [*INPUT_CLASSES, attrs.get("class", "") if attrs else ""]
                ),
            }
        )


class UnfoldAdminIntegerRangeWidget(MultiWidget):
    template_name = "unfold/widgets/range.html"

    def __init__(self, attrs: dict[str, Any] | None = None) -> None:
        if attrs is None:
            attrs = {}

        attrs["class"] = " ".join(
            [*INPUT_CLASSES, attrs.get("class", "") if attrs else ""]
        )

        _widgets = (NumberInput(attrs=attrs), NumberInput(attrs=attrs))

        super().__init__(_widgets, attrs)

    def decompress(self, value: str | None) -> tuple[Callable | None, ...]:
        if value:
            return value.lower, value.upper
        return None, None


class UnfoldAdminEmailInputWidget(AdminEmailInputWidget):
    def __init__(self, attrs: dict[str, Any] | None = None) -> None:
        super().__init__(
            attrs={
                **(attrs or {}),
                "class": " ".join(
                    [*INPUT_CLASSES, attrs.get("class", "") if attrs else ""]
                ),
            }
        )


class FileFieldMixin:
    def get_context(self, name, value, attrs):
        widget = super().get_context(name, value, attrs)

        widget["widget"].update(
            {
                "class": " ".join([*CHECKBOX_CLASSES, *["form-check-input"]]),
                "file_wrapper_class": " ".join(FILE_CLASSES),
                "file_input_class": " ".join(
                    [
                        self.attrs.get("class", ""),
                        *[
                            "opacity-0",
                            "pointer-events-none",
                        ],
                    ]
                ),
            }
        )

        return widget


class UnfoldAdminImageFieldWidget(FileFieldMixin, AdminFileWidget):
    template_name = "unfold/widgets/clearable_file_input.html"


class UnfoldAdminFileFieldWidget(FileFieldMixin, AdminFileWidget):
    template_name = "unfold/widgets/clearable_file_input_small.html"


class UnfoldAdminImageSmallFieldWidget(FileFieldMixin, AdminFileWidget):
    template_name = "unfold/widgets/clearable_file_input_small.html"


class UnfoldAdminDateWidget(AdminDateWidget):
    template_name = "unfold/widgets/date.html"

    def __init__(
        self, attrs: dict[str, Any] | None = None, format: str | None = None
    ) -> None:
        attrs = {
            **(attrs or {}),
            "class": " ".join(
                [
                    "vDateField",
                    *DATETIME_CLASSES,
                    attrs.get("class", "") if attrs else "",
                ]
            ),
            "size": "10",
        }
        super().__init__(attrs=attrs, format=format)

    class Media:
        js = [
            "admin/js/core.js",
            "admin/js/calendar.js",
            "admin/js/admin/DateTimeShortcuts.js",
        ]


class UnfoldAdminSingleDateWidget(AdminDateWidget):
    template_name = "unfold/widgets/date.html"

    def __init__(
        self, attrs: dict[str, Any] | None = None, format: str | None = None
    ) -> None:
        attrs = {
            **(attrs or {}),
            "class": " ".join(
                [
                    "vDateField",
                    *DATETIME_CLASSES,
                    attrs.get("class", "") if attrs else "",
                ]
            ),
            "size": "10",
        }
        super().__init__(attrs=attrs, format=format)


class UnfoldAdminTimeWidget(AdminTimeWidget):
    template_name = "unfold/widgets/time.html"

    def __init__(
        self, attrs: dict[str, Any] | None = None, format: str | None = None
    ) -> None:
        attrs = {
            **(attrs or {}),
            "class": " ".join(
                [
                    "vTimeField",
                    *DATETIME_CLASSES,
                    attrs.get("class", "") if attrs else "",
                ]
            ),
            "size": "8",
        }
        super().__init__(attrs=attrs, format=format)

    class Media:
        js = [
            "admin/js/core.js",
            "admin/js/calendar.js",
            "admin/js/admin/DateTimeShortcuts.js",
        ]


class UnfoldAdminSingleTimeWidget(AdminTimeWidget):
    template_name = "unfold/widgets/time.html"

    def __init__(
        self, attrs: dict[str, Any] | None = None, format: str | None = None
    ) -> None:
        attrs = {
            **(attrs or {}),
            "class": " ".join(
                [
                    "vTimeField",
                    *DATETIME_CLASSES,
                    attrs.get("class", "") if attrs else "",
                ]
            ),
            "size": "8",
        }
        super().__init__(attrs=attrs, format=format)


class UnfoldAdminTextareaWidget(AdminTextareaWidget):
    template_name = "unfold/widgets/textarea.html"

    def __init__(self, attrs: dict[str, Any] | None = None) -> None:
        attrs = attrs or {}

        super().__init__(
            attrs={
                **(attrs or {}),
                "class": " ".join(
                    [
                        "vLargeTextField",
                        *TEXTAREA_CLASSES,
                        attrs.get("class", "") if attrs else "",
                    ]
                ),
            }
        )


class UnfoldAdminExpandableTextareaWidget(UnfoldAdminTextareaWidget):
    def __init__(self, attrs: dict[str, Any] | None = None) -> None:
        attrs = attrs or {}

        attrs.update({"rows": 2})

        super().__init__(
            attrs={
                **(attrs or {}),
                "class": " ".join(
                    [
                        "vLargeTextField",
                        *TEXTAREA_CLASSES,
                        *TEXTAREA_EXPANDABLE_CLASSES,
                        attrs.get("class", "") if attrs else "",
                    ]
                ),
            }
        )


class UnfoldAdminSplitDateTimeWidget(AdminSplitDateTime):
    template_name = "unfold/widgets/split_datetime.html"

    def __init__(self, attrs: dict[str, Any] | None = None) -> None:
        widgets = [
            UnfoldAdminDateWidget(attrs={"placeholder": _("Date")}),
            UnfoldAdminTimeWidget(attrs={"placeholder": _("Time")}),
        ]
        MultiWidget.__init__(self, widgets, attrs)

    class Media:
        js = [
            "admin/js/core.js",
            "admin/js/calendar.js",
            "admin/js/admin/DateTimeShortcuts.js",
        ]


class UnfoldAdminSplitDateTimeVerticalWidget(AdminSplitDateTime):
    template_name = "unfold/widgets/split_datetime_vertical.html"

    def __init__(
        self,
        attrs: dict[str, Any] | None = None,
        date_attrs: dict[str, Any] | None = None,
        time_attrs: dict[str, Any] | None = None,
        date_label: str | None = None,
        time_label: str | None = None,
    ) -> None:
        self.date_label = date_label
        self.time_label = time_label

        widgets = [
            UnfoldAdminDateWidget(attrs=date_attrs),
            UnfoldAdminTimeWidget(attrs=time_attrs),
        ]
        MultiWidget.__init__(self, widgets, attrs)

    def get_context(
        self, name: str, value: Any, attrs: dict[str, Any] | None
    ) -> dict[str, Any]:
        context = super().get_context(name, value, attrs)

        if self.date_label is not None:
            context["date_label"] = self.date_label
        else:
            context["date_label"] = _("Date")

        if self.time_label is not None:
            context["time_label"] = self.time_label
        else:
            context["time_label"] = _("Time")

        return context


class UnfoldAdminIntegerFieldWidget(AdminIntegerFieldWidget):
    def __init__(self, attrs: dict[str, Any] | None = None) -> None:
        super().__init__(
            attrs={
                **(attrs or {}),
                "class": " ".join(
                    [*INPUT_CLASSES, attrs.get("class", "") if attrs else ""]
                ),
            }
        )


class UnfoldAdminDecimalFieldWidget(AdminIntegerFieldWidget):
    def __init__(self, attrs: dict[str, Any] | None = None) -> None:
        super().__init__(
            attrs={
                **(attrs or {}),
                "class": " ".join(
                    [*INPUT_CLASSES, attrs.get("class", "") if attrs else ""]
                ),
            }
        )


class UnfoldAdminBigIntegerFieldWidget(AdminBigIntegerFieldWidget):
    def __init__(self, attrs: dict[str, Any] | None = None) -> None:
        super().__init__(
            attrs={
                **(attrs or {}),
                "class": " ".join(
                    [*INPUT_CLASSES, attrs.get("class", "") if attrs else ""]
                ),
            }
        )


class UnfoldAdminNullBooleanSelectWidget(NullBooleanSelect):
    template_name = "unfold/widgets/select.html"

    def __init__(self, attrs=None):
        if attrs is None:
            attrs = {}

        attrs["class"] = " ".join(
            [*SELECT_CLASSES, attrs.get("class", "") if attrs else ""]
        )
        super().__init__(attrs)


class UnfoldAdminSelectWidget(Select):
    template_name = "unfold/widgets/select.html"

    def __init__(self, attrs=None, choices=()):
        if attrs is None:
            attrs = {}

        attrs["class"] = " ".join(
            [*SELECT_CLASSES, attrs.get("class", "") if attrs else ""]
        )
        super().__init__(attrs, choices)


class UnfoldAdminSelect2Widget(Select):
    def __init__(self, attrs=None, choices=()):
        if attrs is None:
            attrs = {}

        attrs["data-theme"] = "admin-autocomplete"
        attrs["class"] = "unfold-admin-autocomplete admin-autocomplete"

        super().__init__(attrs, choices)

    class Media:
        js = (
            "admin/js/vendor/jquery/jquery.js",
            "admin/js/vendor/select2/select2.full.js",
            "admin/js/jquery.init.js",
            "unfold/js/select2.init.js",
        )
        css = {
            "screen": (
                "admin/css/vendor/select2/select2.css",
                "admin/css/autocomplete.css",
            ),
        }


class UnfoldAdminSelectMultipleWidget(SelectMultiple):
    def __init__(self, attrs=None, choices=()):
        if attrs is None:
            attrs = {}

        attrs["class"] = " ".join(
            [*SELECT_CLASSES, attrs.get("class", "") if attrs else ""]
        )
        super().__init__(attrs, choices)


class UnfoldAdminSelect2MultipleWidget(SelectMultiple):
    def __init__(self, attrs=None, choices=()):
        if attrs is None:
            attrs = {}

        attrs["data-theme"] = "admin-autocomplete"
        attrs["class"] = "unfold-admin-autocomplete admin-autocomplete"

        super().__init__(attrs, choices)

    class Media:
        js = (
            "admin/js/vendor/jquery/jquery.js",
            "admin/js/vendor/select2/select2.full.js",
            "admin/js/jquery.init.js",
            "unfold/js/select2.init.js",
        )
        css = {
            "screen": (
                "admin/css/vendor/select2/select2.css",
                "admin/css/autocomplete.css",
            ),
        }


class UnfoldAdminRadioSelectWidget(AdminRadioSelect):
    template_name = "unfold/widgets/radio.html"
    option_template_name = "unfold/widgets/radio_option.html"

    def __init__(self, radio_style: int | None = None, *args, **kwargs):
        super().__init__(*args, **kwargs)

        if radio_style is None:
            radio_style = VERTICAL

        self.radio_style = radio_style
        self.attrs["class"] = " ".join([*RADIO_CLASSES, self.attrs.get("class", "")])

    def get_context(self, *args, **kwargs) -> dict[str, Any]:
        context = super().get_context(*args, **kwargs)
        context.update({"radio_style": self.radio_style})
        return context


class UnfoldAdminCheckboxSelectMultiple(CheckboxSelectMultiple):
    template_name = "unfold/widgets/radio.html"
    option_template_name = "unfold/widgets/radio_option.html"

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.attrs = {
            "class": " ".join([*CHECKBOX_CLASSES, self.attrs.get("class", "")])
        }


try:
    from djmoney.forms.widgets import MoneyWidget
    from djmoney.settings import CURRENCY_CHOICES

    class UnfoldAdminMoneyWidget(MoneyWidget):
        template_name = "unfold/widgets/split_money.html"

        def __init__(self, *args, **kwargs):
            if "attrs" in kwargs:
                attrs = kwargs.pop("attrs")
            else:
                attrs = {}

            super().__init__(
                amount_widget=UnfoldAdminTextInputWidget(attrs=attrs),
                currency_widget=UnfoldAdminSelectWidget(
                    choices=CURRENCY_CHOICES,
                    attrs={
                        "aria-label": _("Select currency"),
                    },
                ),
            )

except ImportError:

    class UnfoldAdminMoneyWidget:
        def __init__(self, *args, **kwargs):
            raise UnfoldException("django-money not installed")


class UnfoldBooleanWidget(CheckboxInput):
    def __init__(
        self, attrs: dict[str, Any] | None = None, check_test: Callable = None
    ) -> None:
        if attrs is None:
            attrs = {}

        super().__init__(
            {
                **(attrs or {}),
                "class": " ".join(
                    [*CHECKBOX_CLASSES, attrs.get("class", "") if attrs else ""]
                ),
            },
            check_test,
        )


class UnfoldBooleanSwitchWidget(CheckboxInput):
    def __init__(
        self, attrs: dict[str, Any] | None = None, check_test: Callable = None
    ) -> None:
        super().__init__(
            attrs={
                **(attrs or {}),
                "class": " ".join(
                    [*SWITCH_CLASSES, attrs.get("class", "") if attrs else ""]
                ),
            },
            check_test=None,
        )


class UnfoldRelatedFieldWidgetWrapper(RelatedFieldWidgetWrapper):
    template_name = "unfold/widgets/related_widget_wrapper.html"


class UnfoldForeignKeyRawIdWidget(ForeignKeyRawIdWidget):
    template_name = "unfold/widgets/foreign_key_raw_id.html"

    def __init__(
        self,
        rel: ForeignObjectRel,
        admin_site: AdminSite,
        attrs: dict | None = None,
        using: Any | None = None,
    ) -> None:
        attrs = {
            **(attrs or {}),
            "class": " ".join(
                [
                    "vForeignKeyRawIdAdminField",
                    *INPUT_CLASSES,
                    attrs.get("class", "") if attrs else "",
                ]
            ),
        }
        super().__init__(rel, admin_site, attrs, using)


class UnfoldAdminPasswordInput(PasswordInput):
    def __init__(self, attrs=None, render_value=False):
        super().__init__(
            {
                **(attrs or {}),
                "class": " ".join(
                    [*INPUT_CLASSES, attrs.get("class", "") if attrs else ""]
                ),
            },
            render_value,
        )



from crispy_forms.layout import BaseInput, LayoutObject
from crispy_forms.utils import TEMPLATE_PACK
from django.template.loader import render_to_string

from unfold.widgets import BUTTON_CLASSES


class ButtonClassesMixin:
    def __init__(self, *args, css_class=None, **kwargs):
        classes = BUTTON_CLASSES

        if css_class:
            classes.append(css_class)

        self.field_classes = " ".join(classes)

        super().__init__(*args, **kwargs)


class Submit(ButtonClassesMixin, BaseInput):
    input_type = "submit"


class Button(ButtonClassesMixin, BaseInput):
    input_type = "button"


class FieldsetSubheader(LayoutObject):
    template = "unfold_crispy/layout/fieldset_subheader.html"

    def __init__(self, title=None, *args, **kwargs):
        self.title = title
        super().__init__(*args, **kwargs)

    def render(self, form, context, template_pack=TEMPLATE_PACK, **kwargs):
        return render_to_string(
            self.template,
            {
                "title": self.title,
            },
        )


class Hr(LayoutObject):
    template = "unfold_crispy/layout/hr.html"

    def __init__(self, title=None, *args, **kwargs):
        self.title = title
        super().__init__(*args, **kwargs)

    def render(self, form, context, template_pack=TEMPLATE_PACK, **kwargs):
        return render_to_string(
            self.template,
            {
                "title": self.title,
            },
        )



import json
from collections.abc import Iterable, Mapping
from typing import Any

from django import template
from django.contrib.admin.helpers import AdminForm, Fieldset
from django.contrib.admin.views.main import PAGE_VAR, ChangeList
from django.contrib.admin.widgets import RelatedFieldWidgetWrapper
from django.core.paginator import Paginator
from django.db.models import Model
from django.db.models.options import Options
from django.forms import BoundField, CheckboxSelectMultiple, Field
from django.http import HttpRequest, QueryDict
from django.template import Context, Library, Node, RequestContext, TemplateSyntaxError
from django.template.base import NodeList, Parser, Token, token_kwargs
from django.template.loader import render_to_string
from django.urls import reverse_lazy
from django.utils.safestring import SafeText, mark_safe
from django.utils.translation import gettext_lazy as _

from unfold.components import ComponentRegistry
from unfold.dataclasses import UnfoldAction
from unfold.enums import ActionVariant
from unfold.widgets import UnfoldAdminMoneyWidget, UnfoldAdminSplitDateTimeWidget

register = Library()


def _get_tabs_list(
    context: RequestContext, page: str, opts: Options | None = None
) -> list:
    tabs_list = []
    page_id = None

    if page not in ["changeform", "changelist"]:
        page_id = page

    for tab in context.get("tab_list", []):
        if page_id:
            if tab.get("page") == page_id:
                tabs_list = tab["items"]
                break

            continue

        if "models" not in tab:
            continue

        for tab_model in tab["models"]:
            if isinstance(tab_model, str):
                if str(opts) == tab_model and page == "changelist":
                    tabs_list = tab["items"]
                    break
            elif isinstance(tab_model, dict) and str(opts) == tab_model["name"]:
                is_detail = tab_model.get("detail", False)

                if (page == "changeform" and is_detail) or (
                    page == "changelist" and not is_detail
                ):
                    tabs_list = tab["items"]
                    break
    return tabs_list


@register.simple_tag(name="tab_list", takes_context=True)
def tab_list(context: RequestContext, page: str, opts: Options | None = None) -> str:
    inlines_list = []
    datasets_list = []
    data = {
        "nav_global": context.get("nav_global"),
        "actions_detail": context.get("actions_detail"),
        "actions_list": context.get("actions_list"),
        "actions_items": context.get("actions_items"),
        "is_popup": context.get("is_popup"),
        "tabs_list": _get_tabs_list(context, page, opts),
    }

    # If the changeform is rendered and there are no custom tab navigation
    # specified, check for inlines to put into tabs
    if page == "changeform" and len(data["tabs_list"]) == 0:
        for inline in context.get("inline_admin_formsets", []):
            if opts and hasattr(inline.opts, "tab"):
                inlines_list.append(inline)

        if len(inlines_list) > 0:
            data["inlines_list"] = inlines_list

        for dataset in context.get("datasets", []):
            if dataset and hasattr(dataset, "tab"):
                datasets_list.append(dataset)

        if len(datasets_list) > 0:
            data["datasets_list"] = datasets_list

    return render_to_string(
        "unfold/helpers/tab_list.html",
        request=context["request"],
        context=data,
    )


@register.simple_tag(name="render_section", takes_context=True)
def render_section(context: Context, section_class, instance: Model) -> str:
    return section_class(context.request, instance).render()


@register.simple_tag(name="has_nav_item_active")
def has_nav_item_active(items: list) -> bool:
    for item in items:
        if "active" in item and item["active"]:
            return True

    return False


@register.filter
def class_name(value: Any) -> str:
    return value.__class__.__name__


@register.filter
def is_list(value: Any) -> str:
    return isinstance(value, list)


@register.filter
def has_active_item(items: list[dict]) -> bool:
    for item in items:
        if "active" in item and item["active"]:
            return True

    return False


@register.filter
def index(indexable: Mapping[int, Any], i: int) -> Any:
    try:
        return indexable[i]
    except (KeyError, TypeError):
        return None


@register.filter
def tabs(adminform: AdminForm) -> list[Fieldset]:
    result = []

    for fieldset in adminform:
        if "tab" in fieldset.classes and fieldset.name:
            result.append(fieldset)

    return result


class CaptureNode(Node):
    def __init__(self, nodelist: NodeList, varname: str, silent: bool) -> None:
        self.nodelist = nodelist
        self.varname = varname
        self.silent = silent

    def render(self, context: dict[str, Any]) -> str | SafeText:
        output = self.nodelist.render(context)
        context[self.varname] = output
        if self.silent:
            return ""
        else:
            return output


@register.tag(name="capture")
def do_capture(parser: Parser, token: Token) -> CaptureNode:
    """
    Capture the contents of a tag output.
    Usage:
    .. code-block:: html+django
        {% capture %}..{% endcapture %}                    # output in {{ capture }}
        {% capture silent %}..{% endcapture %}             # output in {{ capture }} only
        {% capture as varname %}..{% endcapture %}         # output in {{ varname }}
        {% capture as varname silent %}..{% endcapture %}  # output in {{ varname }} only
    For example:
    .. code-block:: html+django
        {# Allow templates to override the page title/description #}
        <meta name="description" content="{% capture as meta_description %}
            {% block meta-description %}{% endblock %}{% endcapture %}" />
        <title>{% capture as meta_title %}{% block meta-title %}Untitled{% endblock %}{% endcapture %}</title>
        {# copy the values to the Social Media meta tags #}
        <meta property="og:description" content="{% block og-description %}{{ meta_description }}{% endblock %}" />
        <meta name="twitter:title" content="{% block twitter-title %}{{ meta_title }}{% endblock %}" />
    """
    bits = token.split_contents()

    # tokens
    t_as = "as"
    t_silent = "silent"
    var = "capture"
    silent = False

    num_bits = len(bits)
    if len(bits) > 4:
        raise TemplateSyntaxError(
            "'capture' node supports '[as variable] [silent]' parameters."
        )
    elif num_bits == 4:
        t_name, t_as, var, t_silent = bits
        silent = True
    elif num_bits == 3:
        t_name, t_as, var = bits
    elif num_bits == 2:
        t_name, t_silent = bits
        silent = True
    else:
        var = "capture"
        silent = False

    if t_silent != "silent" or t_as != "as":
        raise TemplateSyntaxError(
            "'capture' node expects 'as variable' or 'silent' syntax."
        )

    nodelist = parser.parse(("endcapture",))
    parser.delete_first_token()
    return CaptureNode(nodelist, var, silent)


class RenderComponentNode(template.Node):
    def __init__(
        self,
        template_name: str,
        nodelist: NodeList,
        extra_context: dict | None = None,
        include_context: bool = False,
        *args,
        **kwargs,
    ):
        self.template_name = template_name
        self.nodelist = nodelist
        self.extra_context = extra_context or {}
        self.include_context = include_context
        super().__init__(*args, **kwargs)

    def render(self, context: RequestContext) -> str:
        values = {
            name: var.resolve(context) for name, var in self.extra_context.items()
        }

        if "component_class" in values:
            values = ComponentRegistry.create_instance(
                values["component_class"],
                request=context.request if hasattr(context, "request") else None,
            ).get_context_data(**values)

        context_copy = context.new()
        context_copy.update(context.flatten())
        context_copy.update(values)
        children = self.nodelist.render(context_copy)

        if len(children) > 0:
            values.update(
                {
                    "children": children,
                }
            )

        if self.include_context:
            values.update(context.flatten())

        return render_to_string(
            self.template_name,
            request=context.request if hasattr(context, "request") else None,
            context=values,
        )


@register.tag("component")
def do_component(parser: Parser, token: Token) -> str:
    bits = token.split_contents()

    if len(bits) < 2:
        raise TemplateSyntaxError(
            f"{bits[0]} tag takes at least one argument: the name of the template to be included."
        )

    options = {}
    remaining_bits = bits[2:]

    while remaining_bits:
        option = remaining_bits.pop(0)

        if option in options:
            raise TemplateSyntaxError(
                f"The {option} option was specified more than once."
            )

        if option == "with":
            value = token_kwargs(remaining_bits, parser, support_legacy=False)

            if not value:
                raise TemplateSyntaxError(
                    '"with" in {bits[0]} tag needs at least one keyword argument.'
                )
        elif option == "include_context":
            value = True
        else:
            raise TemplateSyntaxError(f"Unknown argument for {bits[0]} tag: {option}.")

        options[option] = value

    include_context = options.get("include_context", False)
    nodelist = parser.parse(("endcomponent",))
    template_name = bits[1][1:-1]

    extra_context = options.get("with", {})
    parser.next_token()

    return RenderComponentNode(template_name, nodelist, extra_context, include_context)


@register.filter
def add_css_class(field: Field, classes: list | tuple) -> Field:
    if type(classes) in (list, tuple):
        classes = " ".join(classes)

    if "class" in field.field.widget.attrs:
        field.field.widget.attrs["class"] += f" {classes}"
    else:
        field.field.widget.attrs["class"] = classes

    return field


@register.inclusion_tag(
    "unfold/templatetags/preserve_changelist_filters.html",
    takes_context=True,
    name="preserve_filters",
)
def preserve_changelist_filters(context: Context) -> dict[str, dict[str, str]]:
    """
    Generate hidden input fields to preserve filters for POST forms.
    """
    request: HttpRequest | None = context.get("request")
    changelist: ChangeList | None = context.get("cl")

    if not request or not changelist:
        return {"params": {}}

    used_params: set[str] = {
        param for spec in changelist.filter_specs for param in spec.used_parameters
    }
    preserved_params: dict[str, str] = {
        param: value for param, value in request.GET.items() if param not in used_params
    }

    return {"params": preserved_params}


@register.simple_tag(takes_context=True)
def element_classes(context: Context, key: str) -> str:
    if key in context.get("element_classes", {}):
        if isinstance(context["element_classes"][key], list | tuple):
            return " ".join(context["element_classes"][key])

        return context["element_classes"][key]

    return ""


@register.simple_tag(takes_context=True)
def fieldset_rows_classes(context: Context) -> str:
    classes = [
        "aligned",
    ]

    if not context.get("stacked"):
        classes.extend(
            [
                "border",
                "border-base-200",
                "rounded-default",
                "shadow-xs",
                "dark:border-base-800",
            ]
        )

    return " ".join(set(classes))


@register.simple_tag(takes_context=True)
def fieldset_row_classes(context: Context) -> str:
    classes = [
        "form-row",
        "field-row",
        "group/row",
    ]

    formset = context.get("inline_admin_formset", None)
    line = context.get("line")

    # Hide the field in case of ordering field for sorting
    for field in line:
        if (
            formset
            and hasattr(field.field, "name")
            and field.field.name == getattr(formset.opts, "ordering_field", None)
            and getattr(formset.opts, "hide_ordering_field", False)
        ):
            classes.append("hidden")

    if len(line.fields) > 1:
        classes.extend(
            [
                "grid",
                f"lg:grid-cols-{len(line.fields)}",
            ]
        )

    if not line.has_visible_field:
        classes.append("hidden")

    return " ".join(set(classes))


@register.simple_tag(takes_context=True)
def fieldset_line_classes(context: Context) -> str:
    classes = [
        "field-line",
        "flex",
        "flex-col",
        "grow",
        "group",
        "group/line",
        "px-3",
        "py-2.5",
    ]
    field = context.get("field")
    adminform = context.get("adminform")

    if hasattr(field.field, "name") and field.field.name:
        classes.append(f"field-{field.field.name}")

    if hasattr(field, "errors") and field.errors():
        classes.append("errors")

    if (
        adminform
        and hasattr(adminform.model_admin, "compressed_fields")
        and adminform.model_admin.compressed_fields
    ):
        classes.extend(
            [
                "border-b",
                "border-base-200",
                "border-dashed",
                "group-[.last]/row:border-b-0",
                "lg:border-l",
                "lg:flex-row",
                "dark:border-base-800",
                "lg:first:border-l-0",
            ]
        )

    return " ".join(set(classes))


@register.simple_tag(takes_context=True)
def action_item_classes(context: Context, action: UnfoldAction) -> str:
    classes = [
        "border",
        "border-base-200",
        "max-md:-mt-px",
        "max-md:first:rounded-t-default",
        "max-md:last:rounded-b-default",
        "md:-ml-px",
        "md:first:rounded-l-default",
        "md:last:rounded-r-default",
    ]

    if "variant" not in action:
        variant = ActionVariant.DEFAULT
    else:
        variant = action["variant"]

    if variant == ActionVariant.PRIMARY:
        classes.extend(
            [
                "border-primary-700",
                "bg-primary-600",
                "text-white",
                "dark:border-primary-500",
            ]
        )
    elif variant == ActionVariant.DANGER:
        classes.extend(
            [
                "border-red-700",
                "bg-red-600",
                "text-white",
                "dark:border-red-500",
            ]
        )
    elif variant == ActionVariant.SUCCESS:
        classes.extend(
            [
                "border-green-700",
                "bg-green-600",
                "text-white",
                "dark:border-green-500",
            ]
        )
    elif variant == ActionVariant.INFO:
        classes.extend(
            [
                "border-blue-700",
                "bg-blue-600",
                "text-white",
                "dark:border-blue-500",
            ]
        )
    elif variant == ActionVariant.WARNING:
        classes.extend(
            [
                "border-orange-700",
                "bg-orange-600",
                "text-white",
                "dark:border-orange-500",
            ]
        )
    else:
        classes.extend(
            [
                "border-base-200",
                "hover:text-primary-600",
                "dark:hover:text-primary-500",
                "dark:border-base-700",
            ]
        )

    return " ".join(set(classes))


@register.filter
def changeform_data(adminform: AdminForm) -> str:
    fields = {}

    for fieldset in adminform:
        for line in fieldset:
            for field in line:
                if isinstance(field.field, dict):
                    continue

                if isinstance(
                    field.field.field.widget, UnfoldAdminSplitDateTimeWidget
                ) or isinstance(field.field.field.widget, UnfoldAdminMoneyWidget):
                    for index, _widget in enumerate(field.field.field.widget.widgets):
                        fields[
                            f"{field.field.name}{field.field.field.widget.widgets_names[index]}"
                        ] = None
                elif isinstance(field.field.field.widget, CheckboxSelectMultiple):
                    fields[field.field.name] = []
                else:
                    fields[field.field.name] = None

    return mark_safe(json.dumps(fields))


@register.filter
def changeform_condition(field: BoundField) -> BoundField:
    if isinstance(field.field, dict):
        return field

    if isinstance(field.field.field.widget, RelatedFieldWidgetWrapper):
        field.field.field.widget.widget.attrs["x-model.fill"] = field.field.name
        field.field.field.widget.widget.attrs["x-init"] = mark_safe(
            f"const $ = django.jQuery; $(function () {{ const select = $('#{field.field.auto_id}'); select.on('change', (ev) => {{ {field.field.name} = select.val(); }}); }});"
        )
    elif isinstance(
        field.field.field.widget, UnfoldAdminSplitDateTimeWidget
    ) or isinstance(field.field.field.widget, UnfoldAdminMoneyWidget):
        for index, widget in enumerate(field.field.field.widget.widgets):
            field_name = (
                f"{field.field.name}{field.field.field.widget.widgets_names[index]}"
            )

            widget.attrs["x-model.fill"] = field_name
    else:
        field.field.field.widget.attrs["x-model.fill"] = field.field.name

    return field


@register.simple_tag
def infinite_paginator_url(cl, i):
    return cl.get_query_string({PAGE_VAR: i})


@register.simple_tag
def elided_page_range(paginator: Paginator, number: int) -> list[int | str] | None:
    if not paginator or not number:
        return None

    return paginator.get_elided_page_range(number=number)


@register.simple_tag(takes_context=True)
def querystring_params(
    context: RequestContext, query_key: str, query_value: str
) -> str:
    request = context.get("request")
    result = QueryDict(mutable=True)

    for key, values in request.GET.lists():
        if key == query_key:
            continue

        for value in values:
            result[key] = value

    result[query_key] = query_value

    return result.urlencode()


@register.simple_tag(name="unfold_querystring", takes_context=True)
def unfold_querystring(context, *args, **kwargs):
    """
    Duplicated querystring template tag from Django core to allow
    it using in Django 4.x. Once 4.x is not supported, remove it.
    """
    if not args:
        args = [context.request.GET]
    params = QueryDict(mutable=True)
    for d in [*args, kwargs]:
        if not isinstance(d, Mapping):
            raise TemplateSyntaxError(
                "querystring requires mappings for positional arguments (got "
                f"{d!r} instead)."
            )
        for key, value in d.items():
            if not isinstance(key, str):
                raise TemplateSyntaxError(
                    f"querystring requires strings for mapping keys (got {key!r} "
                    "instead)."
                )
            if value is None:
                params.pop(key, None)
            elif isinstance(value, Iterable) and not isinstance(value, str):
                params.setlist(key, value)
            else:
                params[key] = value
    query_string = params.urlencode() if params else ""
    return f"?{query_string}"


@register.simple_tag(takes_context=True)
def header_title(context: RequestContext) -> str:
    parts = []
    opts = context.get("opts")
    current_app = (
        context.request.current_app
        if hasattr(context.request, "current_app")
        else "admin"
    )

    if opts:
        parts.append(
            {
                "link": reverse_lazy(f"{current_app}:app_list", args=[opts.app_label]),
                "title": opts.app_config.verbose_name,
            }
        )

        if (original := context.get("original")) and not isinstance(original, str):
            parts.append(
                {
                    "link": reverse_lazy(
                        f"{current_app}:{original._meta.app_label}_{original._meta.model_name}_changelist"
                    ),
                    "title": original._meta.verbose_name_plural,
                }
            )

            parts.append(
                {
                    "link": reverse_lazy(
                        f"{current_app}:{original._meta.app_label}_{original._meta.model_name}_change",
                        args=[original.pk],
                    ),
                    "title": original,
                }
            )
        elif object := context.get("object"):
            parts.append(
                {
                    "link": reverse_lazy(
                        f"{current_app}:{object._meta.app_label}_{object._meta.model_name}_changelist"
                    ),
                    "title": object._meta.verbose_name_plural,
                }
            )

            parts.append(
                {
                    "link": reverse_lazy(
                        f"{current_app}:{object._meta.app_label}_{object._meta.model_name}_change",
                        args=[object.pk],
                    ),
                    "title": object,
                }
            )
        else:
            parts.append(
                {
                    "link": reverse_lazy(
                        f"{current_app}:{opts.app_label}_{opts.model_name}_changelist"
                    ),
                    "title": opts.verbose_name_plural,
                }
            )
    elif object := context.get("object"):
        parts.append(
            {
                "link": reverse_lazy(
                    f"{current_app}:app_list", args=[object._meta.app_label]
                ),
                "title": object._meta.app_label,
            }
        )

        parts.append(
            {
                "link": reverse_lazy(
                    f"{current_app}:{object._meta.app_label}_{object._meta.model_name}_changelist",
                ),
                "title": object._meta.verbose_name_plural,
            }
        )

        parts.append(
            {
                "link": reverse_lazy(
                    f"{current_app}:{object._meta.app_label}_{object._meta.model_name}_change",
                    args=[object.pk],
                ),
                "title": object,
            }
        )
    elif (model_admin := context.get("model_admin")) and hasattr(model_admin, "model"):
        parts.append(
            {
                "link": reverse_lazy(
                    f"{current_app}:app_list", args=[model_admin.model._meta.app_label]
                ),
                "title": model_admin.model._meta.app_label,
            }
        )

        parts.append(
            {
                "link": reverse_lazy(
                    f"{current_app}:{model_admin.model._meta.app_label}_{model_admin.model._meta.model_name}_changelist",
                ),
                "title": model_admin.model._meta.verbose_name_plural,
            }
        )

    if not opts and (content_title := context.get("content_title")):
        parts.append(
            {
                "title": content_title,
            }
        )

    if len(parts) == 0:
        username = (
            context.request.user.get_short_name() or context.request.user.get_username()
        )
        parts.append({"title": f"{_('Welcome')} {username}"})

    return render_to_string(
        "unfold/helpers/header_title.html",
        request=context.request,
        context={
            "parts": parts,
        },
    )


@register.simple_tag(takes_context=True)
def admin_object_app_url(context: RequestContext, object: Model, arg: str) -> str:
    current_app = (
        context.request.current_app
        if hasattr(context.request, "current_app")
        else "admin"
    )

    return f"{current_app}:{object._meta.app_label}_{object._meta.model_name}_{arg}"


@register.filter
def has_nested_tables(table: dict) -> bool:
    return any(
        isinstance(row, dict) and "table" in row for row in table.get("rows", [])
    )
